


/*
 *
 *          Copyright (C) 1997, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1997. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */





#include "base/date.h"
#include "base/timeofda.h"

#include "ui/psdev.h"
#include "ui/font.h"
#include "ui/arc.h"
#include "ui/chord.h"
#include "ui/piewedge.h"
#include "ui/ellipse.h"
#include <iostream.h>




#if defined(__GNUC__) && __GNUC_MINOR__ >= 6
template class CL_Binding0<UI_PostscriptDevice>;
#elif defined(_MSC_VER)
template CL_Binding0<UI_PostscriptDevice>;
#endif

typedef CL_Binding0<UI_PostscriptDevice> PSDevBind;




static const double
    PointsPerInch = 72.24,
    MMPerInch     = 25.4;

UI_PostscriptDevice::UI_PostscriptDevice (ostream& strm,
                                          double pw,
                                          double ph,
                                          const char* title,
                                          const char* creator)
    : _stream (strm), _widthPts (pw * PointsPerInch),
      _heightPts (ph * PointsPerInch), _pageNo (1)
{
    _handle = 0;
    _brush  = new UI_Brush (this, UI_Color (UIColor_White), UIBrush_Solid);
    _pen    = new UI_Pen   (this);
    _font   = new UI_Font (this);
    *_font  = UI_FontDesc ("Times-Roman", 10); // Hard-wire the initial
                                               // setting

    PSDevBind bind1 (this, &UI_PostscriptDevice::_FontChanged);
    _font->AddDependent (bind1);
    
    PSDevBind bind2 (this, &UI_PostscriptDevice::_PenChanged);
    _pen->AddDependent (bind2);
    
    CL_String prolog;
    CL_Date today = CL_Date::Today ();
    CL_TimeOfDay now = CL_TimeOfDay::Now ();
    prolog.AssignWithFormat ("%%!PS-Adobe-2.0\n"
                             "%%%%Title: %s\n"
                             "%%%%Creator: %s\n"
                             "%%%%CreationDate: %s %s\n"
                             "%%%%Pages: (atend)\n"
                             "%%%%EndComments\n"
                             "/YACLdict 250 dict def\nYACLdict begin\n"
                             "%%%%EndProlog\n"
                             "%%%%Page: 1 1\n",
                             title ? title : "YACL Job",
                             creator ? creator : "YACL",
                             today.AsString().AsPtr(),
                             now.AsString().AsPtr()
        );
    _DoOutput (prolog);
    _FontChanged (); // Force initial font output
}


UI_PostscriptDevice::~UI_PostscriptDevice ()
{
    CL_String ostrg;
    ostrg.AssignWithFormat ("\nshowpage\n%%%%Trailer\nend\n"
                            "%%%%Pages: %d\n", _pageNo);
    _DoOutput (ostrg);
}



// 
// --------------------------- Query ---------------------------------
//     

double UI_PostscriptDevice::HorzPixelsPerMM () const
{
    return _widthPts;
}


double UI_PostscriptDevice::VertPixelsPerMM () const
{
    return _heightPts;
}



UI_Rectangle UI_PostscriptDevice::DrawingArea() const
{
    return UI_Rectangle (0, 0, _widthPts, _heightPts);
}


UI_Rectangle UI_PostscriptDevice::DrawingAreaInMM() const
{
    return UI_Rectangle (0, 0, _widthPts / PointsPerInch * MMPerInch,
                         _heightPts / PointsPerInch * MMPerInch);
}


bool UI_PostscriptDevice::StartNewPage ()
{
    CL_String ostrg;
    _pageNo++;
    ostrg.AssignWithFormat ("\nshowpage\n\n%%%%Page: %ld %ld\n",
                            _pageNo, _pageNo);
    _DoOutput (ostrg);
    return TRUE;
}


// 
// -------------------------- Draw  operations ------------------------
// 


void UI_PostscriptDevice::_BeginFigure (const char* figureName)
{
    if (figureName) {
        _DoOutput ("\n%  ");
        _DoOutput (figureName);
    }
    CL_String oStrg;
    _DoOutput ("\nnewpath\n");
    UI_Color penColor = Pen().Color();
    CL_String ostrg;
    oStrg.AssignWithFormat ("%f %f %f setrgbcolor\n", penColor.Red(),
                            penColor.Green(), penColor.Blue());
    _DoOutput (oStrg);
}


void UI_PostscriptDevice::_EndFigure (long opts)
{
    CL_String oStrg;
    if ((opts & UID_Outline) && _pen->Pattern() != UIPen_None)
        _DoOutput ("gsave stroke grestore\n");
    if (opts & UID_Fill) {
        UI_Color brushColor = _brush->Color();
        oStrg.AssignWithFormat ("%f %f %f setrgbcolor\n", brushColor.Red(),
                                brushColor.Green(), brushColor.Blue());
        _DoOutput (oStrg);
        _DoOutput ("gsave fill grestore\n");
    }
}


bool UI_PostscriptDevice::DrawLine (const UI_Point& p, const UI_Point& q)
{
    _BeginFigure ("DrawLine");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    ostrg.AssignWithFormat ("newpath\n"
                            "%ld %ld moveto\n"
                            "%ld %ld lineto\nstroke\n",
                            p.XCoord(), (long) (_heightPts - p.YCoord()),
                            q.XCoord(), (long) (_heightPts - q.YCoord())
                            );
    _DoOutput (ostrg);
    _EndFigure ((long) UID_Outline);
    _DoOutput ("grestore\n\n");
    return TRUE;
}


void UI_PostscriptDevice::DrawRectangle (const UI_Rectangle& rect,
                                         ushort opt)
{
    _BeginFigure ("DrawRectangle");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    UI_Point pt = rect.BottomLeft();
    long w = rect.Width(), h = rect.Height();
    ostrg.AssignWithFormat ("%ld %ld moveto\n"
                            "0 %ld rlineto\n"
                            "%ld 0 rlineto\n"
                            "0 %ld rlineto\n"
                            "%ld 0 rlineto\n",
                            pt.XCoord(), (long) (_heightPts - pt.YCoord()),
                            h-1, w-1, -h+1, -w+1);
    _DoOutput (ostrg);
    _EndFigure (opt);
    _DoOutput ("grestore\n\n");
}


void UI_PostscriptDevice::DrawEllipse (const UI_Rectangle& rect,
                                       ushort opt)
{
    _BeginFigure ("DrawEllipse");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    UI_Point center = rect.Center();
    long w = rect.Width(), h = rect.Height();
    ostrg.AssignWithFormat ("%ld %ld translate\n%f %f scale\n"
                            "0 0 %f 0 360 arc\n",
                            center.XCoord(),
                            (long) (_heightPts - center.YCoord()), // Center
                            1.0, ((float) h) / w,  // Scale factor
                            w/2.0 // radius of curvature
                            );
    _DoOutput (ostrg);
    _EndFigure (opt);
    _DoOutput ("grestore\n\n");
}


void UI_PostscriptDevice::DrawArc (const UI_Arc &arc)
{
    _BeginFigure ("DrawArc");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    UI_Rectangle r = arc.BoundingRectangle ();
    UI_Point center = r.Center();
    long w = r.Width(), h = r.Height();
    ostrg.AssignWithFormat ("%ld %ld translate\n%f %f scale\n"
                            "0 0 %f %f %f arc\n",
                            center.XCoord(),
                            (long) (_heightPts - center.YCoord()), // Center
                            1.0, ((float) h) / w,  // Scale factor
                            w/2.0, // radius of curvature
                            arc.StartAngle()/64.0,
                            (arc.StartAngle() + arc.SubtendedAngle())/64.0
                            );
    _DoOutput (ostrg);
    _EndFigure (UID_Outline);
    _DoOutput ("grestore\n\n");
}


void UI_PostscriptDevice::DrawChord (const UI_Chord& chord,
                                     ushort opt)
{
    _BeginFigure ("DrawChord");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    UI_Rectangle r = chord.Ellipse().BoundingRectangle ();
    UI_Point center = r.Center();
    long w = r.Width(), h = r.Height();
    ostrg.AssignWithFormat ("%ld %ld translate\n%f %f scale\n"
                            "%f %f %f 0 360 arc closepath\n",
                            center.XCoord(),
                            (long) (_heightPts - center.YCoord()), // Center
                            1.0, ((float) h) / w,  // Scale factor
                            chord.StartAngle()/64.0,
                            (chord.StartAngle() + chord.SubtendedAngle())/64.0,
                            w/2.0 // radius of curvature
                            );
    _DoOutput (ostrg);
    _EndFigure (opt);
    _DoOutput ("grestore\n\n");
}


void UI_PostscriptDevice::DrawPieWedge (const UI_PieWedge& pie,
                                        ushort opt)
{
    _BeginFigure ("DrawPieWedge");

    _DoOutput ("gsave\n");
    CL_String ostrg;
    UI_Point center = pie.Center();
    UI_Rectangle r = pie.BoundingRectangle ();
    long w = r.Width(), h = r.Height();
    UI_Point ctr = pie.Center ();
    ostrg.AssignWithFormat ("%ld %ld translate\n%f %f scale\n0 0 moveto\n"
                            "0 0 %f %f %f arc closepath\n",
                            center.XCoord(),
                            (long) (_heightPts - center.YCoord()), // Center
                            1.0, ((float) h) / w,  // Scale factor
                            w/2.0, // radius of curvature
                            pie.StartAngle()/64.0,
                            (pie.StartAngle() + pie.SubtendedAngle())/64.0
                            );
    _DoOutput (ostrg);
    _EndFigure (opt);
    _DoOutput ("grestore\n\n");
}


void UI_PostscriptDevice::DrawPolygon (UI_Point points[], short numPoints,
                                       ushort opt)
{
    _BeginFigure ("DrawPolygon");
    _DoOutput ("gsave\n");
    if (numPoints > 1) {
        CL_String ostrg;
        ostrg.AssignWithFormat ("%ld %ld moveto\n", points[0].XCoord(),
                                points[0].YCoord());
        _DoOutput (ostrg);
        for (short i = 1; i < numPoints; i++) {
            ostrg.AssignWithFormat ("%ld %ld lineto\n", points[i].XCoord(),
                                    points[i].YCoord());
            _DoOutput (ostrg);
        }
        _DoOutput ("closepath");
    }
    _EndFigure (opt);
    _DoOutput ("grestore\n");
}

void UI_PostscriptDevice::DrawPolyLine (UI_Point points[],
                                        short numPoints)
{
    _BeginFigure ("DrawPolygon");
    _DoOutput ("gsave\n");
    if (numPoints > 1) {
        CL_String ostrg;
        ostrg.AssignWithFormat ("%ld %ld moveto\n", points[0].XCoord(),
                                points[0].YCoord());
        _DoOutput (ostrg);
        for (short i = 1; i < numPoints; i++) {
            ostrg.AssignWithFormat ("%ld %ld lineto\n", points[i].XCoord(),
                                    points[i].YCoord());
            _DoOutput (ostrg);
        }
    }
    _EndFigure ((long) UID_Outline);
    _DoOutput ("grestore\n");
}


void UI_PostscriptDevice::DrawPoint (const UI_Point& p)
{
    DrawLine (p, p);
}



void UI_PostscriptDevice::WriteString (const char* str,
                                       const UI_Rectangle& box,
                                       UI_TextStyle style)
{
    CL_String text (str);
    text.ReplaceAll ("(", "\(");
    text.ReplaceAll (")", "\)");
    CL_String oStrg;
    _BeginFigure ("WriteString");
    long x = box.Left(), y = box.Bottom(), w = box.Width(), h = box.Height();
    switch (style) {
    case UIText_Center: {
        oStrg.AssignWithFormat ("%ld %ld moveto\n(%s) dup stringwidth "
                                " 2 div neg exch 2 div neg exch rmoveto\n",
                                x + w/2,
                                (long) (_heightPts - y + h/2), str);
        break;
    }

    case UIText_Right: {
        oStrg.AssignWithFormat ("%ld %ld moveto\n(%s) dup stringwidth "
                                " 2 div neg exch neg exch rmoveto\n",
                                x + w - 1, 
                                (long) (_heightPts - y + h/2), str);
        break;
    }

    default:
        oStrg.AssignWithFormat ("%ld %ld moveto\n(%s)\n",
                                box.Left(),
                                (long) (_heightPts - box.Bottom()), str);
        break;
    };
    _DoOutput (oStrg);
    _DoOutput ("show\n\n");
    // Should also take care of underlining and strike-out
}



bool UI_PostscriptDevice::SaveState()
{
    UI_DrawingSurface::SaveState ();
    _DoOutput ("\ngsave\n");
    return TRUE;
}


bool UI_PostscriptDevice::RestoreState ()
{
    UI_DrawingSurface::RestoreState ();
    _DoOutput ("\ngrestore\n");
    return TRUE;
}


void UI_PostscriptDevice::AppendCode (const char* s)
{
    _DoOutput (s);
}

void UI_PostscriptDevice::_DoOutput (const char* string)
{
    _stream << string;
}




bool UI_PostscriptDevice::_PenChanged ()
{
    CL_String ostrg;
    ostrg.AssignWithFormat ("%ld setlinewidth\n", _pen->Thickness());
    switch (_pen->Pattern()) {
    case UIPen_Solid:
        ostrg += "[] 0 setdash\n";
        break;
        
    case UIPen_Dot:
        ostrg += "[2] 0 setdash\n";
        break;
        
    case UIPen_Dash:
        ostrg += "[6] 0 setdash\n";
        break;

    case UIPen_None:
        break;
    }
    _DoOutput (ostrg);
    return TRUE;
}



bool UI_PostscriptDevice::_FontChanged ()
{
    CL_String oStrg, fontName;
    CL_String typeFace = _font->TypeFace();
    long style = _font->Style();
    const char* styleDesc = "";

    // Determine the style parameters
    if (style & (UIFont_BoldFace | UIFont_Italic))
        styleDesc = "-BoldOblique";
    else if (style & UIFont_BoldFace)
        styleDesc = "-Bold";
    else if (style & UIFont_Italic)
        styleDesc = "-Oblique";

    // Determine the type face
    if (typeFace == UIFont_Courier)
        fontName = "Courier";
    else if (typeFace == UIFont_Times) {
        if (style)
            fontName = "Times";
        else
            fontName = "Times-Roman";
    }
    else if (typeFace == UIFont_Helvetica)
        fontName = "Helvetica";

    fontName += styleDesc;
    
    oStrg.AssignWithFormat ("/%s findfont %d scalefont setfont\n",
                            fontName.AsPtr(), _font->PointSize());
    _DoOutput (oStrg);
    return TRUE;
}



